/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2012 Centre Tecnologic de Telecomunicacions de Catalunya (CTTC)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Manuel Requena <manuel.requena@cttc.es>
 */

#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/internet-module.h"
#include "ns3/mobility-module.h"
#include "ns3/lte-module.h"
#include "ns3/applications-module.h"
#include "ns3/point-to-point-module.h"
#include "ns3/config-store-module.h"
#include "ns3/math.h"
#include "ns3/flow-monitor-helper.h"
#include "params.h"



using namespace ns3;



void
NotifyPduReceived (std::string context, uint16_t a, uint8_t b, uint32_t c)
{
  std::cout << context
            << "PDU received!" 
            << " RNTI: " << a
            << " lcid: " << b
            << " packetsize: " << c << std::endl;
}

void
SetImsiPos (uint32_t imsi, NodeContainer c,Vector v)
{
  for
  (uint8_t k = 0; k<c.GetN ();k++)
  {
    Ptr <Node> nI = c.Get (k);
    if (imsi == nI->GetDevice (0)->GetObject <LteUeNetDevice> ()->GetImsi ())
    {
      nI->GetObject <MobilityModel> ()->SetPosition (v);
    }
  }
}

void
GetNodePos (NodeContainer a)
{
  for
  (uint8_t k = 0; k<a.GetN ();k++)
  {
    uint64_t imsiTmp=0;
    //std::cout << int(k) << "   " << int(a.GetN ()) << std::endl;
    Ptr <Node> n1 = a.Get (k);
    Vector pos = n1->GetObject <MobilityModel> ()->GetPosition ();
    if (n1->GetDevice (0)->GetObject <LteUeNetDevice> ()!=NULL)
    {
      imsiTmp = n1->GetDevice (0)->GetObject <LteUeNetDevice> ()->GetImsi ();
    }
    //std::string ko = to_string (k);
    std::cout << "Node has imsi: " << imsiTmp << " and pos x = " << pos.x << " y = " << pos.y << " z = " << pos.z << std::endl;
    
  }
}

double
Dbm (double w)
{
  double db = 10 * log10 (w) + 30;
  return db;
}

void
SinrLevels (std::string context, uint16_t cellId, uint16_t rnti, double rsrp, double sinr)
{
  //not used
  
}

void
PrintUeInfo (Ptr<Node> n)
{
  uint64_t imsi = n->GetDevice (0)->GetObject <LteUeNetDevice> ()->GetImsi ();
  std::cout << "Imsi nr is: " << imsi << std::endl;
}

void
PrintEnbInfo (Ptr<Node> n)
{
  uint16_t cellID = n->GetDevice (0)->GetObject <LteEnbNetDevice> ()->GetCellId ();
  std::cout << "Cell Id is: " << cellID << std::endl;
}

void
NotifyNewUeContext ()
{
  std::cout << "\nNewUeContext created\n";
}

void 
NotifyConnectionEstablishedUe (std::string context, 
                               uint64_t imsi, 
                               uint16_t cellid, 
                               uint16_t rnti)
{
  std::cout << context 
            << " UE IMSI " << imsi 
            << ": connected to CellId " << cellid 
            << " with RNTI " << rnti 
            << std::endl;
}

void 
NotifyHandoverStartUe (std::string context, 
                       uint64_t imsi, 
                       uint16_t cellid, 
                       uint16_t rnti, 
                       uint16_t targetCellId)
{
  std::cout << context 
            << " UE IMSI " << imsi 
            << ": previously connected to CellId " << cellid 
            << " with RNTI " << rnti 
            << ", doing handover to CellId " << targetCellId 
            << std::endl;
}

void 
NotifyHandoverEndOkUe (std::string context, 
                       uint64_t imsi, 
                       uint16_t cellid, 
                       uint16_t rnti)
{
  std::cout << context 
            << " UE IMSI " << imsi 
            << ": successful handover to CellId " << cellid 
            << " with RNTI " << rnti 
            << std::endl;
}

void 
NotifyConnectionEstablishedEnb (std::string context, 
                                uint64_t imsi, 
                                uint16_t cellid, 
                                uint16_t rnti)
{
  std::cout << context 
            << " eNB CellId " << cellid 
            << ": successful connection of UE with IMSI " << imsi 
            << " RNTI " << rnti 
            << std::endl;
}

void 
NotifyHandoverStartEnb (std::string context, 
                        uint64_t imsi, 
                        uint16_t cellid, 
                        uint16_t rnti, 
                        uint16_t targetCellId)
{
  std::cout << context 
            << " eNB CellId " << cellid 
            << ": start handover of UE with IMSI " << imsi 
            << " RNTI " << rnti 
            << " to CellId " << targetCellId 
            << std::endl;
}

void 
NotifyHandoverEndOkEnb (std::string context, 
                        uint64_t imsi, 
                        uint16_t cellid, 
                        uint16_t rnti)
{
  std::cout << context 
            << " eNB CellId " << cellid 
            << ": completed handover of UE with IMSI " << imsi 
            << " RNTI " << rnti 
            << std::endl;
}
 

static void 
CourseChange (std::string foo, Ptr<const MobilityModel> mobility)
{
  Vector pos = mobility->GetPosition ();
  Vector vel = mobility->GetVelocity ();
  std::cout << Simulator::Now () << ", model=" << mobility << ", POS: x=" << pos.x << ", y=" << pos.y
            << ", z=" << pos.z << "; VEL:" << vel.x << ", y=" << vel.y
            << ", z=" << vel.z << std::endl;
}


// Built on the x2 handover example script in order to set up EPC for handover
NS_LOG_COMPONENT_DEFINE ("X2-Base");
int
main (int argc, char *argv[])
{
  // LogLevel logLevel = (LogLevel)(LOG_PREFIX_FUNC | LOG_PREFIX_TIME | LOG_LEVEL_ALL);

  // LogComponentEnable ("LteHelper", logLevel);
  // LogComponentEnable ("EpcHelper", logLevel);
  // LogComponentEnable ("EpcEnbApplication", logLevel);
  // LogComponentEnable ("EpcX2", logLevel);
  // LogComponentEnable ("EpcSgwPgwApplication", logLevel);

  // LogComponentEnable ("LteEnbRrc", logLevel);
  // LogComponentEnable ("LteEnbNetDevice", logLevel);
  // LogComponentEnable ("LteUeRrc", logLevel);
  // LogComponentEnable ("LteUeNetDevice", logLevel);
  
  
  Ptr <UniformRandomVariable> URV = CreateObject <UniformRandomVariable> ();
  URV->SetAttribute ("Min", DoubleValue (20.0));
  URV->SetAttribute ("Max", DoubleValue (25.0));
  

  
  Config::SetDefault ("ns3::UdpClient::Interval", TimeValue (MilliSeconds(10)));
  Config::SetDefault ("ns3::UdpClient::MaxPackets", UintegerValue(1000000));
  Config::SetDefault ("ns3::LteHelper::UseIdealRrc", BooleanValue(false));
  Config::SetDefault ("ns3::LteEnbRrc::ServingCellHandoverThreshold", UintegerValue (30));
  Config::SetDefault ("ns3::LteEnbRrc::NeighbourCellHandoverOffset", UintegerValue (1));
  Config::SetDefault ("ns3::LteEnbRrc::EventA2Threshold", UintegerValue (32));
  Config::SetDefault ("ns3::LteEnbRrc::EventA4Threshold", UintegerValue (2));
  Config::SetDefault ("ns3::LteEnbRrc::SrsPeriodicity", UintegerValue(160));
  Config::SetDefault ("ns3::LteEnbNetDevice::DlBandwidth", UintegerValue (100));
  
  Config::SetDefault ("ns3::RandomWalk2dMobilityModel::Mode", StringValue ("Time"));
  Config::SetDefault ("ns3::RandomWalk2dMobilityModel::Time", StringValue ("0.2s"));
  Config::SetDefault ("ns3::RandomWalk2dMobilityModel::Speed", StringValue ("ns3::UniformRandomVariable[Min=20.0|Max=25.0]"));
  Config::SetDefault ("ns3::RandomWalk2dMobilityModel::Bounds", StringValue ("0|200|0|200"));
  
  Config::SetDefault ("ns3::RrFfMacScheduler::HarqEnabled", BooleanValue (false));
  //Config::SetDefault ("ns3::LteAmc::AmcModel", StringValue ("PiroEW2010"));
  //Config::SetDefault ("ns3::")

  // Command line arguments
  CommandLine cmd;
  cmd.AddValue("numberOfUes", "Number of UEs", numberOfUes);
  cmd.AddValue("numberOfEnbs", "Number of eNodeBs", numberOfEnbs);
  cmd.AddValue("simTime", "Total duration of the simulation (in seconds)",simTime);
  cmd.Parse(argc, argv);


  Ptr<LteHelper> lteHelper = CreateObject<LteHelper> ();
  Ptr<EpcHelper> epcHelper = CreateObject<EpcHelper> ();
  lteHelper->SetEpcHelper (epcHelper);
  lteHelper->SetSchedulerType("ns3::RrFfMacScheduler");
  

  Ptr<Node> pgw = epcHelper->GetPgwNode ();

  // Create a single RemoteHost
  NodeContainer remoteHostContainer;
  remoteHostContainer.Create (1);
  Ptr<Node> remoteHost = remoteHostContainer.Get (0);
  InternetStackHelper internet;
  internet.Install (remoteHostContainer);

  // Create the Internet
  PointToPointHelper p2ph;
  p2ph.SetDeviceAttribute ("DataRate", DataRateValue (DataRate ("100Gb/s")));
  p2ph.SetDeviceAttribute ("Mtu", UintegerValue (1500));
  p2ph.SetChannelAttribute ("Delay", TimeValue (Seconds (0.010)));
  NetDeviceContainer internetDevices = p2ph.Install (pgw, remoteHost);
  Ipv4AddressHelper ipv4h;
  ipv4h.SetBase ("1.0.0.0", "255.0.0.0");
  Ipv4InterfaceContainer internetIpIfaces = ipv4h.Assign (internetDevices);
  Ipv4Address remoteHostAddr = internetIpIfaces.GetAddress (1);


  // Routing of the Internet Host (towards the LTE network)
  Ipv4StaticRoutingHelper ipv4RoutingHelper;
  Ptr<Ipv4StaticRouting> remoteHostStaticRouting = ipv4RoutingHelper.GetStaticRouting (remoteHost->GetObject<Ipv4> ());
  // interface 0 is localhost, 1 is the p2p device
  remoteHostStaticRouting->AddNetworkRouteTo (Ipv4Address ("7.0.0.0"), Ipv4Mask ("255.0.0.0"), 1);

  NodeContainer ueNodes;
  NodeContainer enbNodes;
  enbNodes.Create(numberOfEnbs);
  ueNodes.Create(numberOfUes);
  
  Ptr <Node> macroNode = enbNodes.Get (0);
  Ptr <Node> picoNode = enbNodes.Get (1);
  Ptr <Node> pUe = ueNodes.Get (0);
 
  /*
  NS_LOG_UNCOND("Mobilitet macro");
  MobilityHelper mobilitetMacro;
  mobilitetMacro.SetMobilityModel("ns3::ConstantPositionMobilityModel");
  mobilitetMacro.Install(macroNode);
  */
  // Mobility model used to test the model
  NS_LOG_UNCOND("Mobility Macro UE");
  MobilityHelper mobilitetMue;
  mobilitetMue.SetPositionAllocator ("ns3::RandomDiscPositionAllocator",
                                 "X", StringValue ("0.0"),
                                 "Y", StringValue ("1000.0"),
                                 "Rho", StringValue ("ns3::UniformRandomVariable[Min=0|Max=1000]"));
  
  mobilitetMue.SetMobilityModel ("ns3::RandomWalk2dMobilityModel",
                                 "Mode", StringValue ("Time"),
                                 "Time", StringValue ("2s"),
                                 "Bounds", StringValue ("-2000|2000|-1000|3000"));
  
  mobilitetMue.Install (ueNodes);
  
  
  
  
  std::cout << "Number of ue nodes: " << ueNodes.GetN () << std::endl;
  
  
  NS_LOG_UNCOND("Mobilitet eNB"); 
  MobilityHelper mobilitetPico;
  mobilitetPico.SetMobilityModel("ns3::ConstantPositionMobilityModel");
  mobilitetPico.Install(enbNodes);
  picoNode->GetObject <MobilityModel> ()->SetPosition (picoPos);
  macroNode->GetObject <MobilityModel> ()->SetPosition (macroPos);
  
  
  

  // Install LTE Devices in eNB and UEs
  NetDeviceContainer enbLteDevs = lteHelper->InstallEnbDevice (enbNodes);
  NetDeviceContainer ueLteDevs = lteHelper->InstallUeDevice (ueNodes);

  // Install the IP stack on the UEs
  internet.Install (ueNodes);
  Ipv4InterfaceContainer ueIpIfaces;
  ueIpIfaces = epcHelper->AssignUeIpv4Address (NetDeviceContainer (ueLteDevs));
  
  
  for (uint32_t u = 0; u < ueNodes.GetN (); ++u)
    {
      Ptr<Node> ueNode = ueNodes.Get (u);
      // Set the default gateway for the UE
      Ptr<Ipv4StaticRouting> ueStaticRouting = ipv4RoutingHelper.GetStaticRouting (ueNode->GetObject<Ipv4> ());
      ueStaticRouting->SetDefaultRoute (epcHelper->GetUeDefaultGatewayAddress (), 1);
    }


  
  //attaching UEs to closest node
  lteHelper->AttachToClosestEnb (ueLteDevs, enbLteDevs);
  
  NS_LOG_LOGIC ("setting up applications");
    
  // Install and start applications on UEs and remote host
  uint16_t dlPort = 10000;
  uint16_t ulPort = 20000;

  // randomize a bit start times to avoid simulation artifacts
  // (e.g., buffer overflows due to packet transmissions happening
  // exactly at the same time) 
  Ptr<UniformRandomVariable> startTimeSeconds = CreateObject<UniformRandomVariable> ();
  startTimeSeconds->SetAttribute ("Min", DoubleValue (0));
  startTimeSeconds->SetAttribute ("Max", DoubleValue (0.05));
     
  for (uint32_t u = 0; u < numberOfUes; ++u)
    {
      Ptr<Node> ue = ueNodes.Get (u);
      // Set the default gateway for the UE
      Ptr<Ipv4StaticRouting> ueStaticRouting = ipv4RoutingHelper.GetStaticRouting (ue->GetObject<Ipv4> ());
      ueStaticRouting->SetDefaultRoute (epcHelper->GetUeDefaultGatewayAddress (), 1);

      for (uint32_t b = 0; b < numBearersPerUe; ++b)
        {
          ++dlPort;
          ++ulPort;

          ApplicationContainer clientApps;
          ApplicationContainer serverApps;
          
          NS_LOG_LOGIC ("installing UDP DL app for UE " << u);
          UdpClientHelper dlClientHelper (ueIpIfaces.GetAddress (u), dlPort);
          clientApps.Add (dlClientHelper.Install (remoteHost));
          PacketSinkHelper dlPacketSinkHelper ("ns3::UdpSocketFactory", 
                                               InetSocketAddress (Ipv4Address::GetAny (), dlPort));
          serverApps.Add (dlPacketSinkHelper.Install (ue));
              
          NS_LOG_LOGIC ("installing UDP UL app for UE " << u);
          UdpClientHelper ulClientHelper (remoteHostAddr, ulPort);
          clientApps.Add (ulClientHelper.Install (ue));
          PacketSinkHelper ulPacketSinkHelper ("ns3::UdpSocketFactory", 
                                               InetSocketAddress (Ipv4Address::GetAny (), ulPort));
          serverApps.Add (ulPacketSinkHelper.Install (remoteHost));  
             
          Ptr<EpcTft> tft = Create<EpcTft> ();
          EpcTft::PacketFilter dlpf;
          dlpf.localPortStart = dlPort;
          dlpf.localPortEnd = dlPort;
          tft->Add (dlpf); 
          EpcTft::PacketFilter ulpf;
          ulpf.remotePortStart = ulPort;
          ulpf.remotePortEnd = ulPort;
          tft->Add (ulpf);
          EpsBearer bearer (EpsBearer::NGBR_VIDEO_TCP_DEFAULT);
          lteHelper->ActivateDedicatedEpsBearer (ueLteDevs.Get (u), bearer, tft);
          
          Time startTime = Seconds (startTimeSeconds->GetValue ());
          serverApps.Start (startTime);
          clientApps.Start (startTime);

        } // end for b
    }


  // Add X2 inteface
  lteHelper->AddX2Interface (enbNodes);

  // X2-based Handover manual
  //lteHelper->HandoverRequest (Seconds (0.200), pUe->GetDevice (0), macroNode->GetDevice (0), picoNode->GetDevice (0));
  
  
  // Uncomment to enable PCAP tracing
  //p2ph.EnablePcapAll("lena-x2-handover");

  //lteHelper->EnableMacTraces ();
  //lteHelper->EnableRlcTraces ();
  //lteHelper->EnablePdcpTraces ();
  //lteHelper->EnablePhyTraces ();
  lteHelper->EnableTraces ();
  
  Ptr<RadioBearerStatsCalculator> rlcStats = lteHelper->GetRlcStats ();
  rlcStats->SetAttribute ("EpochDuration", TimeValue (Seconds (0.05)));
  Ptr<RadioBearerStatsCalculator> pdcpStats = lteHelper->GetPdcpStats ();
  pdcpStats->SetAttribute ("EpochDuration", TimeValue (Seconds (0.05)));
  
  //setting Txpower of different nodes in the HetNet
  Ptr <LteEnbPhy> macroPhy = macroNode->GetDevice (0)->GetObject <LteEnbNetDevice> ()->GetPhy ();
  macroPhy->SetTxPower (macroTx);
  Ptr <LteEnbPhy> picoPhy = picoNode->GetDevice (0)->GetObject <LteEnbNetDevice> ()->GetPhy ();
  picoPhy->SetTxPower (picoTx);
  
  
  // Setting bias CRE values
  
  Ptr <BiasHelper> b = CreateObject <BiasHelper> ();
  b->SetBias (rpB,rqB);
  
  for
  (int enbIt = 0; enbIt<numberOfUes; enbIt++)
  {
    Ptr <Node> n0 = ueNodes.Get (enbIt);
    n0->GetDevice (0)->GetObject <LteUeNetDevice> ()->GetPhy ()->SetBias (b);
  }
  
  b->biasList.push_back (picoNode->GetDevice (0)->GetObject <LteEnbNetDevice> ()->GetCellId ());

  //ABS related
  // Getting Ptrs to rrFfMacScheduler to initialise for ABS
  Ptr<FfMacScheduler> rrFfMacScheduler = macroNode->GetDevice (0)->GetObject <LteEnbNetDevice> ()->GetFfMacSchedule ();
  Ptr<FfMacScheduler> rrFfMacSchedulerPico = picoNode->GetDevice (0)->GetObject <LteEnbNetDevice> ()->GetFfMacSchedule ();
  
  // Get access to physical layer in scheduler
  rrFfMacSchedulerPico->GetObject <RrFfMacScheduler> ()->SetPhy (picoNode->GetDevice (0)->GetObject <LteEnbNetDevice> ()->GetPhy ());
  rrFfMacScheduler->GetObject <RrFfMacScheduler> ()->SetPhy (macroNode->GetDevice (0)->GetObject <LteEnbNetDevice> ()->GetPhy ());
  
  //setting parameters for macro
  rrFfMacScheduler->GetObject <RrFfMacScheduler> ()->SetAbsParameters(absPeriod,absInterval);
  rrFfMacScheduler->GetObject <RrFfMacScheduler> ()->SetMacroNode(true);
  
  //setting parameters for pico
  rrFfMacSchedulerPico->GetObject <RrFfMacScheduler> ()->SetMacroNode (false);
  rrFfMacSchedulerPico->GetObject <RrFfMacScheduler> ()->SetAbsParameters (absPeriod,absInterval);
  
  //setting neighbour links for communication
  rrFfMacScheduler->GetObject <RrFfMacScheduler> ()->SetNeigh (rrFfMacSchedulerPico->GetObject <RrFfMacScheduler> ());
  rrFfMacSchedulerPico->GetObject <RrFfMacScheduler> ()->SetNeigh (rrFfMacScheduler->GetObject <RrFfMacScheduler> ());
  
  
  //Monitoring packet trafic
  FlowMonitorHelper flow;
  Ptr <FlowMonitor> monitor = flow.Install (ueNodes);
  monitor = flow.GetMonitor ();
  
  uint8_t dlBw = macroNode->GetDevice (0)->GetObject <LteEnbNetDevice> ()->GetDlBandwidth ();
  std::cout << "Downlink bandwidth er: " << int(dlBw) << "ressursblokker." << std::endl;
  
  //Mobility trace
  /*
  AsciiTraceHelper ascii;
  MobilityHelper::EnableAsciiAll (ascii.CreateFileStream ("mobility-trace.mob"));
  */
  // connect custom trace sinks for RRC connection establishment and handover notification
  Config::Connect ("/NodeList/*/DeviceList/*/LteEnbRrc/ConnectionEstablished",
                   MakeCallback (&NotifyConnectionEstablishedEnb));
  Config::Connect ("/NodeList/*/DeviceList/*/LteUeRrc/ConnectionEstablished",
                   MakeCallback (&NotifyConnectionEstablishedUe));
  Config::Connect ("/NodeList/*/DeviceList/*/LteEnbRrc/HandoverStart",
                   MakeCallback (&NotifyHandoverStartEnb));
  Config::Connect ("/NodeList/*/DeviceList/*/LteUeRrc/HandoverStart",
                   MakeCallback (&NotifyHandoverStartUe));
  Config::Connect ("/NodeList/*/DeviceList/*/LteEnbRrc/HandoverEndOk",
                   MakeCallback (&NotifyHandoverEndOkEnb));
  Config::Connect ("/NodeList/*/DeviceList/*/LteUeRrc/HandoverEndOk",
                   MakeCallback (&NotifyHandoverEndOkUe));
  //Config::Connect ("/NodeList/*/DeviceList/*/LteEnbRrc/NewUeContext", MakeCallback (&NotifyNewUeContext));
  Config::Connect ("/NodeList/4/DeviceList/0/LteUePhy/ReportCurrentCellRsrpSinr", MakeCallback (&SinrLevels));
  
  Config::Connect ("/NodeList/*/$ns3::MobilityModel/CourseChange",
                   MakeCallback (&CourseChange));
  //Config::Connect ("/NodeList/*/DeviceList/*/LteUeRrc/LteRlc/TxPDU",
  //                 MakeCallback (&NotifyPduReceived));
                   
                   
  
  std::cout << "Simulation starts" << std::endl;
  Simulator::Stop(Seconds(simTime));
  Simulator::Run();
  
  //Flowmonitor
  monitor->SerializeToXmlFile("flowen.flow", true, true);
  
  // GtkConfigStore config;
  // config.ConfigureAttributes();

  Simulator::Destroy();
  
  
  
  return 0;
  

}


